#include <InterfaceKit.h>

// some prototypes for your own use:

bool floyd_8_8(uchar *in_data, rgb_color *in_palette, int in_width, int in_height, BBitmap *dest_bitmap);
bool floyd_32_8(ulong *in_data, int in_width, int in_height, BBitmap *dest_bitmap);

bool floyd_8_8(uchar *in_data, rgb_color *in_palette, int in_width, int in_height, BBitmap *dest_bitmap)
/*
 *	Does a floyd alike dithering to a BBitmap
 * 
 *	in_data		= input data in 8bit format
 *	in_palette	= pointer to a palette with 256 entries
 *	in_width	= width of the picture (== bytes per row!)
 *	in_height	= height of the picture
 *	dest_bitmap	= pointer to a BBitmap with COLOR_8_BIT colorspace
 *  			  Make sure that the BBitmap is big enough...
 *
 *  corrections are done this way:
 *
 *	+-----+-----+- -
 *  |     | 0.5 |
 *  +-----+-----+- -
 *  | 0.5 | 0.0 |
 *  +-----+-----+- -
 *  |     |     |
 *
 *  This allows to use only shifting and avoiding real divisions.
 *  You will note a slight diagonal dithering, but this is no
 *  real disadvantage.
 *
 */
{
	if(dest_bitmap->ColorSpace()!=B_COLOR_8_BIT)	return FALSE;

	// We need much table stuff...

	char 		*delta	= new char[(in_width+1)*2*3];
	uchar		*dst	= (uchar *)dest_bitmap->Bits();
	ulong 		*ipal	= (ulong *)in_palette;
	uchar		*xpal	= system_colors()->index_map;
	ulong		*cpal	= (ulong *)system_colors()->color_list;
	char		*fcwrite= delta+((in_width+1)*3);
	char		*cread	= delta;
	char		*cwrite	= fcwrite;
	ulong		xmod	= dest_bitmap->BytesPerRow();
	long		switcher= 0;

	memset(delta,0,(in_width+1)*2*3);
	
	for(int y=0;y<in_height;y++)
	{
		long rdlast	= 0;
		long gdlast	= 0;
		long bdlast	= 0;
		
		for(int x=0;x<in_width;x++)
		{
			// get 1 pixel from the input buffer and convert it to RGBA form.

			ulong 	p  = ipal[*in_data++];

			// add corrections from previous pixels.
			
			long	r  = ((p>>24)&0xFF) + ((*cread++ + rdlast)/2);
			long	g  = ((p>>16)&0xFF) + ((*cread++ + gdlast)/2);
			long	b  = ((p>> 8)&0xFF) + ((*cread++ + bdlast)/2);

			// fix high and low values

			r = r > 255 ? 255 : r;
			r = r <   0 ?   0 : r;
			g = g > 255 ? 255 : g;
			g = g <   0 ?   0 : g;
			b = b > 255 ? 255 : b;
			b = b <   0 ?   0 : b;

			// lets see which color we get from the BeOS palette.
			
			uchar 	q = xpal[((r<<7)&0x7C00)|((g<<2)&0x03E0)|((b>>3)&0x001F)];
			ulong 	s = cpal[q];

			// output the pixel

			dst[x] = q;

			// save the differences to the requested color

			*cwrite++ = rdlast = r - ((s>>24)&0xFF);
			*cwrite++ = gdlast = g - ((s>>16)&0xFF);
			*cwrite++ = bdlast = b - ((s>> 8)&0xFF);
		}

		// switch correction buffers

		if(switcher)
		{
			cwrite	= delta;
			cread	= fcwrite;
			switcher= 0;
		}
		else
		{
			cread	= delta;
			cwrite	= fcwrite;
			switcher= 1;
		}
		dst+=xmod;
	}
	delete [] delta;
	
	return TRUE;
}

bool floyd_32_8(ulong *in_data, int in_width, int in_height, BBitmap *dest_bitmap)
/*
 *	Does a floyd alike dithering to a BBitmap
 * 
 *	in_data		= input data in RGBA format
 *	in_width	= width of the picture (== bytes per row!)
 *	in_height	= height of the picture
 *	dest_bitmap	= pointer to a BBitmap with COLOR_8_BIT colorspace
 *  			  Make sure that the BBitmap is big enough...
 *
 *  corrections are done this way:
 *
 *	+-----+-----+- -
 *  |     | 0.5 |
 *  +-----+-----+- -
 *  | 0.5 | 0.0 |
 *  +-----+-----+- -
 *  |     |     |
 *
 *  This allows to use only shifting and avoiding real divisions.
 *  You will note a slight diagonal dithering, but this is no
 *  real disadvantage.
 *
 */
{
	if(dest_bitmap->ColorSpace()!=B_COLOR_8_BIT)	return FALSE;

	char 		*delta	= new char[(in_width+1)*2*3];
	uchar		*dst	= (uchar *)dest_bitmap->Bits();
	uchar		*xpal	= system_colors()->index_map;
	ulong		*cpal	= (ulong *)system_colors()->color_list;
	char		*fcwrite= delta+((in_width+1)*3);
	char		*cread	= delta;
	char		*cwrite	= fcwrite;
	ulong		xmod	= dest_bitmap->BytesPerRow();
	long		switcher= 0;

	memset(delta,0,(in_width+1)*2*3);
	
	for(int y=0;y<in_height;y++)
	{
		long rdlast	= 0;
		long gdlast	= 0;
		long bdlast	= 0;
		
		for(int x=0;x<in_width;x++)
		{
			// get 1 pixel from the input buffer

			ulong 	p  = *in_data++;			

			// add corrections from previous pixels.
			
			long	r  = ((p>>24)&0xFF) + ((*cread++ + rdlast)/2);
			long	g  = ((p>>16)&0xFF) + ((*cread++ + gdlast)/2);
			long	b  = ((p>> 8)&0xFF) + ((*cread++ + bdlast)/2);

			// fix high and low values

			r = r > 255 ? 255 : r;
			r = r <   0 ?   0 : r;
			g = g > 255 ? 255 : g;
			g = g <   0 ?   0 : g;
			b = b > 255 ? 255 : b;
			b = b <   0 ?   0 : b;

			// lets see which color we get from the BeOS palette.
			
			uchar 	q = xpal[((r<<7)&0x7C00)|((g<<2)&0x03E0)|((b>>3)&0x001F)];
			ulong 	s = cpal[q];

			// output the pixel

			dst[x] = q;

			// save the differences to the requested color

			*cwrite++ = rdlast = r - ((s>>24)&0xFF);
			*cwrite++ = gdlast = g - ((s>>16)&0xFF);
			*cwrite++ = bdlast = b - ((s>> 8)&0xFF);
		}

		if(switcher)
		{
			cwrite	= delta;
			cread	= fcwrite;
			switcher= 0;
		}
		else
		{
			cread	= delta;
			cwrite	= fcwrite;
			switcher= 1;
		}				
		dst+=xmod;
	}
	delete [] delta;
	
	return TRUE;
}
